#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This script will convert all the steps into CLI callables.

You should not edit this script.
"""

import inspect
import logging
from unittest import mock

import fire

from actk import steps
from actk.bin.all import All

###############################################################################

log = logging.getLogger()
logging.basicConfig(
    level=logging.INFO, format="[%(levelname)4s:%(lineno)4s %(asctime)s] %(message)s"
)

###############################################################################


def cli():
    step_map = {
        name.lower(): step
        for name, step in inspect.getmembers(steps)
        if inspect.isclass(step)
    }

    # Interrupt fire print return
    with mock.patch("fire.core._PrintResult"):
        fire.Fire({**step_map, "all": All})


if __name__ == "__main__":
    cli()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib import Path
from typing import List, Optional, Union

import dask.dataframe as dd
import pandas as pd

from .. import exceptions

#######################################################################################


def check_required_fields(
    dataset: Union[str, Path, pd.DataFrame, dd.DataFrame], required_fields: List[str],
) -> Optional[exceptions.MissingDataError]:
    # Handle dataset provided as string or path
    if isinstance(dataset, (str, Path)):
        dataset = Path(dataset).expanduser().resolve(strict=True)

        # Read dataset
        dataset = dd.read_csv(dataset)

    # Check that all columns provided as required are in the dataset
    missing_fields = set(required_fields) - set(dataset.columns)
    if len(missing_fields) > 0:
        raise exceptions.MissingDataError(dataset, missing_fields)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class Channels:
    NucleusSegmentation = "nucleus_segmentation"
    MembraneSegmentation = "membrane_segmentation"
    DNA = "dna"
    Membrane = "membrane"
    Structure = "structure"
    Brightfield = "brightfield"
    DefaultOrderList = [
        NucleusSegmentation,
        MembraneSegmentation,
        DNA,
        Membrane,
        Structure,
        Brightfield,
    ]


class DatasetFields:
    CellId = "CellId"
    CellIndex = "CellIndex"
    FOVId = "FOVId"
    SourceReadPath = "SourceReadPath"
    NucleusSegmentationReadPath = "NucleusSegmentationReadPath"
    MembraneSegmentationReadPath = "MembraneSegmentationReadPath"
    ChannelIndexDNA = "ChannelIndexDNA"
    ChannelIndexMembrane = "ChannelIndexMembrane"
    ChannelIndexStructure = "ChannelIndexStructure"
    ChannelIndexBrightfield = "ChannelIndexBrightfield"
    StandardizedFOVPath = "StandardizedFOVPath"
    CellFeaturesPath = "CellFeaturesPath"
    CellImage3DPath = "CellImage3DPath"
    CellImage2DAllProjectionsPath = "CellImage2DAllProjectionsPath"
    CellImage2DYXProjectionPath = "CellImage2DYXProjectionPath"
    DiagnosticSheetPath = "DiagnosticSheetPath"
    AllExpectedInputs = [
        CellId,
        CellIndex,
        FOVId,
        SourceReadPath,
        NucleusSegmentationReadPath,
        MembraneSegmentationReadPath,
        ChannelIndexDNA,
        ChannelIndexMembrane,
        ChannelIndexStructure,
        ChannelIndexBrightfield,
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
from pathlib import Path
from typing import List, NamedTuple, Optional, Union

import aicsimageio
import dask.dataframe as dd
import numpy as np
import pandas as pd
import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import matplotlib

from aics_dask_utils import DistributedHandler
from datastep import Step, log_run_params

from ...constants import DatasetFields
from ...utils import dataset_utils
from ..single_cell_images import SingleCellImages

plt.style.use("dark_background")

###############################################################################

log = logging.getLogger(__name__)

###############################################################################

REQUIRED_DATASET_FIELDS = [
    DatasetFields.CellId,
    DatasetFields.CellImage2DAllProjectionsPath,
]


class DiagnosticSheetResult(NamedTuple):
    cell_id: Union[int, str]
    save_path: Optional[Path] = None


class DiagnosticSheetError(NamedTuple):
    cell_id: Union[int, str]
    error: str


def flatten(x):
    result = []
    for el in x:
        if hasattr(el, "__iter__") and not isinstance(el, str):
            result.extend(flatten(el))
        else:
            result.append(el)
    return result


###############################################################################


class MakeDiagnosticSheet(Step):
    def __init__(
        self,
        direct_upstream_tasks: List["Step"] = [SingleCellImages],
        filepath_columns=[DatasetFields.DiagnosticSheetPath],
    ):
        super().__init__(
            direct_upstream_tasks=direct_upstream_tasks,
            filepath_columns=filepath_columns,
        )

    @staticmethod
    def _save_single_figure(
        dataset: pd.DataFrame,
        figure_number: Union[int, str],
        number_of_subplots: int,
        index: int,
        this_metadata: str,
        metadata: Union[str, list],
        multiple_paths: bool,
        figure_path: Union[str, Path],
        feature: Optional[str] = None,
        fig_width: Optional[int] = None,
        fig_height: Optional[int] = None,
    ):
        # Get rows and columns of figure
        columns = int(np.sqrt(number_of_subplots) + 0.5)
        if np.sqrt(number_of_subplots) != int(np.sqrt(number_of_subplots)):
            rows = columns + 1
        else:
            rows = columns

        # Set figure size
        if not fig_width:
            fig_width = columns * 7
        if not fig_height:
            fig_height = rows * 5

        # Set subplots
        fig, ax_array = plt.subplots(
            rows, columns, squeeze=False, figsize=(fig_height, fig_width)
        )

        # Set title
        fig.suptitle(f"{this_metadata}:{figure_number}")

        # Second index to check subplots
        index2 = 0

        for k, ax_row in enumerate(ax_array):
            for j, axes in enumerate(ax_row):
                if index2 < number_of_subplots:
                    # Load feature to plot if feature
                    if feature:
                        with open(dataset[DatasetFields.CellFeaturesPath][index]) as f:
                            this_cell_features = json.load(f)
                        title = "CellId: {0}, {1} {2}: {3}".format(
                            dataset[DatasetFields.CellId][index],
                            "\n",
                            feature,
                            this_cell_features[feature],
                        )
                        axes.set_title(title)
                    else:
                        axes.set_title(
                            f"CellID: {dataset[DatasetFields.CellId][index]}"
                        )
                    axes.axis("off")
                    # Read AllProjections Image
                    img = mpimg.imread(
                        dataset[DatasetFields.CellImage2DAllProjectionsPath][index]
                    )
                    axes.imshow(img)
                    axes.set_aspect(1)

                    # Update fig save path in dataset
                    # If multiple paths, choose the correct index for
                    # figure path
                    if this_metadata != metadata and multiple_paths:
                        dataset.loc[index, DatasetFields.DiagnosticSheetPath][
                            metadata.index(this_metadata)
                        ] = str(figure_path)
                    else:
                        dataset.loc[index, DatasetFields.DiagnosticSheetPath] = str(
                            figure_path
                        )

                    index += 1
                    index2 += 1
                else:
                    axes.axis("off")

        # Savefig
        fig.savefig(figure_path)

        return index, dataset

    def _make_group_plot(
        self,
        dataset: pd.DataFrame,
        this_metadata: str,
        metadata: Union[list, str],
        max_cells: int = 1000,
        feature: Optional[str] = None,
        fig_width: Optional[int] = None,
        fig_height: Optional[int] = None,
    ):
        # Get figure numbers, subplot numbers, and figure paths
        # Number of unique metadata values
        figure_numbers = dataset[this_metadata].value_counts().index.to_numpy()
        # Number of cells that have each unique metadata value
        subplot_numbers = dataset[this_metadata].value_counts().values
        # Paths to save
        figure_paths = []
        for fig_num in figure_numbers:
            figure_paths.append(
                dataset.loc[dataset[this_metadata] == fig_num][
                    DatasetFields.DiagnosticSheetPath
                ].values
            )

        # Index to help loop through figure numbers
        index = 0

        # Set font size
        font = {"weight": "bold", "size": 11}
        matplotlib.rc("font", **font)

        # Loop through figure numbers
        for i in range(len(figure_numbers)):

            this_figure_number = figure_numbers[i]
            this_figure_subplots = subplot_numbers[i]
            # figure_paths[i] is a list of fig paths for all cell IDs
            # that share the same metadata value. Lets choose
            # figure_paths[i][0] as one of those paths (all are same)
            this_figure_path = figure_paths[i][0]

            # Get correct figure path if multiple metadata, i.e.
            # len(figure_paths[i][0]) > 1
            if this_metadata != metadata:
                if isinstance(this_figure_path, list) and len(this_figure_path) > 1:
                    multiple_paths = True
                    this_figure_path = this_figure_path[metadata.index(this_metadata)]
                else:
                    multiple_paths = False

            if this_figure_subplots < max_cells:

                index, dataset = self._save_single_figure(
                    dataset,
                    this_figure_number,
                    this_figure_subplots,
                    index,
                    this_metadata,
                    metadata,
                    multiple_paths,
                    this_figure_path,
                    feature,
                    fig_width,
                    fig_height,
                )
            else:
                # If subplot numbers are more than a max cell number (1000),
                # then make separate figures
                subplot_split = [
                    max_cells for i in range(int(this_figure_subplots / max_cells))
                ] + [
                    (
                        this_figure_subplots
                        - max_cells * int(this_figure_subplots / max_cells)
                    )
                ]

                # pop last element if 0 (figure subplots =
                # max_cells * int(subplots/max_cells))
                if subplot_split[-1] == 0:
                    subplot_split.pop()

                for j, sublot in enumerate(subplot_split):
                    # If split, add a "_{number}" before ".png" in path
                    split_fig_path = Path(
                        str(this_figure_path)[:-4] + "_" + str(j) + ".png"
                    )
                    log.info(f"Splitting this figure and saving to: {split_fig_path}")

                    index, dataset = self._save_single_figure(
                        dataset,
                        this_figure_number,
                        sublot,
                        index,
                        this_metadata,
                        metadata,
                        multiple_paths,
                        split_fig_path,
                        feature,
                        fig_width,
                        fig_height,
                    )

            log.info(
                f"Completed diagnostic sheet for: {this_metadata} {this_figure_number} "
            )

        return dataset

    @staticmethod
    def _collect_group(
        row_index: int,
        row: pd.Series,
        diagnostic_sheet_dir: Path,
        overwrite: bool,
        metadata: str,
    ) -> Union[DiagnosticSheetResult, DiagnosticSheetError]:
        # Don't use dask for image reading
        aicsimageio.use_dask(False)

        try:
            # Get the ultimate end save paths for grouped plot
            if row[str(metadata)] or row[str(metadata)] == 0:
                assert DatasetFields.CellImage2DAllProjectionsPath in row.index
                save_path = (
                    diagnostic_sheet_dir / f"{metadata}-{row[str(metadata)]}.png"
                )
                log.info(
                    f"Generating diagnostic sheet path for cell ID: {row.CellId},"
                    f"{metadata} {row[str(metadata)]}"
                )
            else:
                # else no path to save
                save_path = None

            # Check skip
            if not overwrite and save_path.is_file():
                log.info(
                    f"Skipping diagnostic sheet path for cell ID: {row.CellId},"
                    f"{metadata} {row[str(metadata)]}"
                )
                return DiagnosticSheetResult(row.CellId, None)

            # Return ready to save image
            return DiagnosticSheetResult(row.CellId, str(save_path))
        # Catch and return error
        except Exception as e:
            log.info(
                f"Failed to retrieve the CellImage2DAllProjectionsPath"
                f"for cell ID: {row.CellId},"
                f"{metadata} {row[str(metadata)]}"
                f"Error: {e}"
            )
            return DiagnosticSheetError(row.CellId, str(e))

    @log_run_params
    def run(
        self,
        dataset: Union[str, Path, pd.DataFrame, dd.DataFrame],
        max_cells: int = 1000,
        metadata: Optional[Union[list, str]] = None,
        feature: Optional[str] = None,
        fig_width: Optional[int] = None,
        fig_height: Optional[int] = None,
        distributed_executor_address: Optional[str] = None,
        batch_size: Optional[int] = None,
        overwrite: bool = False,
        **kwargs,
    ):
        """
        Provided a dataset of single cell all projection images, generate a diagnostic
        sheet grouped by desired metadata and feature

        Parameters
        ----------
        dataset: Union[str, Path, pd.DataFrame, dd.DataFrame]
            The primary cell dataset to use for generating 
            diagnistic sheet for a group of cells.

            **Required dataset columns:** *["CellId", "CellImage2DAllProjectionsPath"]*

        max_cells: int
            The maximum number of cells to display on a single diagnostic sheet.
            Deafult: 1000

        metadata: Optional[Union[list, str]]
            The metadata to group cells and generate a diagnostic sheet. 
            For example, "FOVId" or "["FOVId", "ProteinDisplayName"]"

        feature: Optional[str]
            The name of the single cell feature to display. For example, "imsize_orig"

        fig_width: Optional[int]
            Width of the diagnostic sheet figure      
        fig_height: Optional[int]
            Height of the diagnostic sheet figure

        distributed_executor_address: Optional[str]
            An optional executor address to pass to some computation engine.
            Default: None

        batch_size: Optional[int]
            An optional batch size to process n features at a time.
            Default: None (Process all at once)

        overwrite: bool
            If this step has already partially or completely run, should it overwrite
            the previous files or not.
            Default: False (Do not overwrite or regenerate files)

        Returns
        -------
        manifest_save_path: Path
            Path to the produced manifest with the DiagnosticSheetPath column added.
        """
        if isinstance(dataset, (str, Path)):
            dataset = Path(dataset).expanduser().resolve(strict=True)

            # Read dataset
            dataset = pd.read_csv(dataset)

        # Check dataset and manifest have required fields
        dataset_utils.check_required_fields(
            dataset=dataset, required_fields=REQUIRED_DATASET_FIELDS,
        )

        # Create save directories
        diagnostic_sheet_dir = self.step_local_staging_dir
        diagnostic_sheet_dir.mkdir(exist_ok=True)

        # Create empty manifest
        manifest = {
            "Metadata": [],
            "MetadataValue": [],
            DatasetFields.DiagnosticSheetPath: [],
        }

        # Check for metadata
        if metadata:
            # Make metadata a list
            metadata = metadata if isinstance(metadata, list) else [metadata]

            # Process each row
            for j, this_metadata in enumerate(metadata):

                # Create empty diagnostic sheet result dataset and errors
                diagnostic_sheet_result_dataset = []
                errors = []

                with DistributedHandler(distributed_executor_address) as handler:
                    # Start processing
                    diagnostic_sheet_result = handler.batched_map(
                        self._collect_group,
                        # Convert dataframe iterrows into two lists of items to iterate
                        # One list will be row index
                        # One list will be the pandas series of every row
                        *zip(*list(dataset.iterrows())),
                        [diagnostic_sheet_dir for i in range(len(dataset))],
                        [overwrite for i in range(len(dataset))],
                        [this_metadata for i in range(len(dataset))],
                        batch_size=batch_size,
                    )
                # Generate diagnostic sheet dataset rows
                for r in diagnostic_sheet_result:
                    if isinstance(r, DiagnosticSheetResult):
                        diagnostic_sheet_result_dataset.append(
                            {
                                DatasetFields.CellId: r.cell_id,
                                DatasetFields.DiagnosticSheetPath: r.save_path,
                            }
                        )
                    else:
                        errors.append(
                            {DatasetFields.CellId: r.cell_id, "Error": r.error}
                        )

                # Convert diagnostic sheet paths rows to dataframe
                diagnostic_sheet_result_dataset = pd.DataFrame(
                    diagnostic_sheet_result_dataset
                )

                # Drop the various diagnostic sheet columns if they already exist
                # Check at j = 0 because the path will exist at j > 1 if
                # multiple metadata
                drop_columns = []
                if DatasetFields.DiagnosticSheetPath in dataset.columns and j == 0:
                    drop_columns.append(DatasetFields.DiagnosticSheetPath)

                dataset = dataset.drop(columns=drop_columns)

                # Update manifest with these paths if there is data
                if len(diagnostic_sheet_result_dataset) > 0:

                    # Join original dataset to the fov paths
                    dataset = dataset.merge(
                        diagnostic_sheet_result_dataset, on=DatasetFields.CellId
                    )

                    # If j > 0 (i.e. multiple metadata), we will append new paths
                    # to the same DiagnosticSheetPath column
                    if j > 0:
                        path_list = dataset[
                            [
                                DatasetFields.DiagnosticSheetPath + "_x",
                                DatasetFields.DiagnosticSheetPath + "_y",
                            ]
                        ].values.tolist()

                        # Flatten pathlist
                        path_list = [
                            list(flatten(this_path)) for this_path in path_list
                        ]
                        dataset[DatasetFields.DiagnosticSheetPath] = path_list

                        # Delete the _x and _y columns that are auto generated
                        del dataset[DatasetFields.DiagnosticSheetPath + "_x"]
                        del dataset[DatasetFields.DiagnosticSheetPath + "_y"]

                if j == 0:
                    dataset.dropna().reset_index(inplace=True)

                # Call the group plotting function
                dataset = self._make_group_plot(
                    dataset,
                    this_metadata,
                    metadata,
                    max_cells,
                    feature,
                    fig_width,
                    fig_height,
                )

                for cell_index, row in dataset.iterrows():
                    if j > 0:
                        this_path = row[DatasetFields.DiagnosticSheetPath][
                            metadata.index(this_metadata)
                        ]
                    else:
                        this_path = row[DatasetFields.DiagnosticSheetPath]

                    if this_path not in manifest[DatasetFields.DiagnosticSheetPath]:
                        manifest["Metadata"].append(this_metadata)
                        manifest["MetadataValue"].append(row[this_metadata])
                        manifest[DatasetFields.DiagnosticSheetPath].append(this_path)

                # Save errored cells to JSON
                with open(
                    self.step_local_staging_dir / "errors.json", "w"
                ) as write_out:
                    json.dump(errors, write_out)

            self.manifest = pd.DataFrame(manifest)
        else:
            # If no metadata, just return input manifest
            self.manifest = dataset

        # Save manifest to CSV
        manifest_save_path = self.step_local_staging_dir / "manifest.csv"
        self.manifest.to_csv(manifest_save_path, index=False)

        return manifest_save_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
from pathlib import Path
from typing import NamedTuple, Optional, Union

import aicsimageio
import dask.dataframe as dd
import pandas as pd
from aics_dask_utils import DistributedHandler
from aicsimageio import AICSImage
from datastep import Step, log_run_params

from ...constants import DatasetFields
from ...utils import dataset_utils, image_utils
from ..standardize_fov_array import StandardizeFOVArray

###############################################################################

log = logging.getLogger(__name__)

###############################################################################

REQUIRED_DATASET_FIELDS = [
    DatasetFields.CellId,
    DatasetFields.CellIndex,
    DatasetFields.FOVId,
    DatasetFields.StandardizedFOVPath,
]


class SingleCellFeaturesResult(NamedTuple):
    cell_id: Union[int, str]
    path: Path


class SingleCellFeaturesError(NamedTuple):
    cell_id: int
    error: str


###############################################################################


class SingleCellFeatures(Step):
    def __init__(
        self,
        direct_upstream_tasks=[StandardizeFOVArray],
        filepath_columns=[DatasetFields.CellFeaturesPath],
    ):
        super().__init__(
            direct_upstream_tasks=direct_upstream_tasks,
            filepath_columns=filepath_columns,
        )

    @staticmethod
    def _generate_single_cell_features(
        row_index: int,
        row: pd.Series,
        cell_ceiling_adjustment: int,
        save_dir: Path,
        overwrite: bool,
    ) -> Union[SingleCellFeaturesResult, SingleCellFeaturesError]:
        # Don't use dask for image reading
        aicsimageio.use_dask(False)

        # Get the ultimate end save path for this cell
        save_path = save_dir / f"{row.CellId}.json"

        # Check skip
        if not overwrite and save_path.is_file():
            log.info(f"Skipping cell feature generation for Cell Id: {row.CellId}")
            return SingleCellFeaturesResult(row.CellId, save_path)

        # Overwrite or didn't exist
        log.info(f"Beginning cell feature generation for CellId: {row.CellId}")

        # Wrap errors for debugging later
        try:
            # Read the standardized FOV
            image = AICSImage(row.StandardizedFOVPath)

            # Preload image data
            image.data

            # Select and adjust cell shape ceiling for this cell
            adjusted = image_utils.select_and_adjust_segmentation_ceiling(
                image=image.get_image_data("CYXZ", S=0, T=0),
                cell_index=row.CellIndex,
                cell_ceiling_adjustment=cell_ceiling_adjustment,
            )

            # Crop the FOV to the segmentation portions
            cropped = image_utils.crop_raw_channels_with_segmentation(
                image=adjusted, channels=image.get_channel_names(),
            )

            # Generate features
            features = image_utils.get_features_from_image(cropped)

            # Save to JSON
            with open(save_path, "w") as write_out:
                json.dump(features, write_out)

            log.info(f"Completed cell feature generation for CellId: {row.CellId}")
            return SingleCellFeaturesResult(row.CellId, save_path)

        # Catch and return error
        except Exception as e:
            log.info(
                f"Failed cell feature generation for CellId: {row.CellId}. Error: {e}"
            )
            return SingleCellFeaturesError(row.CellId, str(e))

    @log_run_params
    def run(
        self,
        dataset: Union[str, Path, pd.DataFrame, dd.DataFrame],
        cell_ceiling_adjustment: int = 7,
        distributed_executor_address: Optional[str] = None,
        batch_size: Optional[int] = None,
        overwrite: bool = False,
        **kwargs,
    ):
        """
        Provided a dataset generate a features JSON file for each cell.

        Parameters
        ----------
        dataset: Union[str, Path, pd.DataFrame, dd.DataFrame]
            The primary cell dataset to use for generating features JSON for each cell.

            **Required dataset columns:** *["CellId", "CellIndex", "FOVId",
            "StandardizedFOVPath"]*

        cell_ceiling_adjustment: int
            The adjust to use for raising the cell shape ceiling. If <= 0, this will be
            ignored and cell data will be selected but not adjusted.
            Default: 7

        distributed_executor_address: Optional[str]
            An optional executor address to pass to some computation engine.
            Default: None

        batch_size: Optional[int]
            An optional batch size to process n features at a time.
            Default: None (Process all at once)

        overwrite: bool
            If this step has already partially or completely run, should it overwrite
            the previous files or not.
            Default: False (Do not overwrite or regenerate files)

        Returns
        -------
        manifest_save_path: Path
            Path to the produced manifest with the CellFeaturesPath column added.
        """
        # Handle dataset provided as string or path
        if isinstance(dataset, (str, Path)):
            dataset = Path(dataset).expanduser().resolve(strict=True)

            # Check dataset and manifest have required fields
            dataset_utils.check_required_fields(
                dataset=dataset, required_fields=REQUIRED_DATASET_FIELDS,
            )

        # Read dataset
        dataset = pd.read_csv(dataset)

        # Create features directory
        features_dir = self.step_local_staging_dir / "cell_features"
        features_dir.mkdir(exist_ok=True)

        # Process each row
        with DistributedHandler(distributed_executor_address) as handler:
            # Start processing
            results = handler.batched_map(
                self._generate_single_cell_features,
                # Convert dataframe iterrows into two lists of items to iterate over
                # One list will be row index
                # One list will be the pandas series of every row
                *zip(*list(dataset.iterrows())),
                # Pass the other parameters as list of the same thing for each
                # mapped function call
                [cell_ceiling_adjustment for i in range(len(dataset))],
                [features_dir for i in range(len(dataset))],
                [overwrite for i in range(len(dataset))],
                batch_size=batch_size,
            )

        # Generate features paths rows
        cell_features_dataset = []
        errors = []
        for result in results:
            if isinstance(result, SingleCellFeaturesResult):
                cell_features_dataset.append(
                    {
                        DatasetFields.CellId: result.cell_id,
                        DatasetFields.CellFeaturesPath: result.path,
                    }
                )
            else:
                errors.append(
                    {DatasetFields.CellId: result.cell_id, "Error": result.error}
                )

        # Convert features paths rows to dataframe
        cell_features_dataset = pd.DataFrame(cell_features_dataset)

        # Drop CellFeaturesPath column if it already exists
        if DatasetFields.CellFeaturesPath in dataset.columns:
            dataset = dataset.drop(columns=[DatasetFields.CellFeaturesPath])

        # Join original dataset to the fov paths
        self.manifest = dataset.merge(cell_features_dataset, on=DatasetFields.CellId)

        # Save manifest to CSV
        manifest_save_path = self.step_local_staging_dir / "manifest.csv"
        self.manifest.to_csv(manifest_save_path, index=False)

        # Save errored cells to JSON
        with open(self.step_local_staging_dir / "errors.json", "w") as write_out:
            json.dump(errors, write_out)

        return manifest_save_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import logging
from pathlib import Path
from typing import List, NamedTuple, Optional, Union

import aicsimageio
import aicsimageprocessing as proc
import dask.dataframe as dd
import numpy as np
import pandas as pd
from aics_dask_utils import DistributedHandler
from aicsimageio import AICSImage, transforms
from aicsimageio.writers import OmeTiffWriter
from datastep import Step, log_run_params
from imageio import imwrite

from ...constants import Channels, DatasetFields
from ...utils import dataset_utils, image_utils
from ..single_cell_features import SingleCellFeatures

###############################################################################

log = logging.getLogger(__name__)

###############################################################################

REQUIRED_DATASET_FIELDS = [
    DatasetFields.CellId,
    DatasetFields.StandardizedFOVPath,
    DatasetFields.CellFeaturesPath,
]


class CellImagesResult(NamedTuple):
    cell_id: Union[int, str]
    path_3d: Path
    path_2d_all_proj: Path
    path_2d_yx_proj: Path


class CellImagesError(NamedTuple):
    cell_id: Union[int, str]
    error: str


###############################################################################


class SingleCellImages(Step):
    def __init__(
        self,
        direct_upstream_tasks=[SingleCellFeatures],
        filepath_columns=[
            DatasetFields.CellImage3DPath,
            DatasetFields.CellImage2DAllProjectionsPath,
            DatasetFields.CellImage2DYXProjectionPath,
        ],
    ):
        super().__init__(
            direct_upstream_tasks=direct_upstream_tasks,
            filepath_columns=filepath_columns,
        )

    @staticmethod
    def _get_registered_image_size(row_index: int, row: pd.Series) -> List[int]:
        # Open cell features JSON
        with open(row.CellFeaturesPath, "r") as read_in:
            cell_features = json.load(read_in)

        # Return registered image size
        return cell_features["imsize_registered"]

    @staticmethod
    def _generate_single_cell_images(
        row_index: int,
        row: pd.Series,
        cell_ceiling_adjustment: int,
        bounding_box: np.ndarray,
        projection_method: str,
        cell_images_3d_dir: Path,
        cell_images_2d_all_proj_dir: Path,
        cell_images_2d_yx_proj_dir: Path,
        overwrite: bool,
    ) -> Union[CellImagesResult, CellImagesError]:
        # Don't use dask for image reading
        aicsimageio.use_dask(False)

        # Get the ultimate end save paths for this cell
        cell_image_3d_save_path = cell_images_3d_dir / f"{row.CellId}.ome.tiff"
        cell_image_2d_all_proj_save_path = (
            cell_images_2d_all_proj_dir / f"{row.CellId}.png"
        )
        cell_image_2d_yx_proj_save_path = (
            cell_images_2d_yx_proj_dir / f"{row.CellId}.png"
        )

        # Check skip
        if (
            not overwrite
            # Only skip if all images exist for this cell
            and all(
                p.is_file()
                for p in [
                    cell_image_3d_save_path,
                    cell_image_2d_all_proj_save_path,
                    cell_image_2d_yx_proj_save_path,
                ]
            )
        ):
            log.info(f"Skipping single cell image generation for CellId: {row.CellId}")
            return CellImagesResult(
                row.CellId,
                cell_image_3d_save_path,
                cell_image_2d_all_proj_save_path,
                cell_image_2d_yx_proj_save_path,
            )

        # Overwrite or didn't exist
        log.info(f"Beginning single cell image generation for CellId: {row.CellId}")

        # Wrap errors for debugging later
        try:
            # Initialize image object with standardized FOV
            standardized_image = AICSImage(row.StandardizedFOVPath)
            channels = standardized_image.get_channel_names()

            # Preload image data
            standardized_image.data

            # Select and adjust cell shape ceiling for this cell
            image = image_utils.select_and_adjust_segmentation_ceiling(
                # Unlike most other operations, we can read in normal "CZYX" dimension
                # order here as all future operations are expecting it
                image=standardized_image.get_image_data("CYXZ", S=0, T=0),
                cell_index=row.CellIndex,
                cell_ceiling_adjustment=cell_ceiling_adjustment,
            )

            # Perform a rigid registration on the image
            image, _, _ = proc.cell_rigid_registration(
                image,
                # Reorder bounding box as image is currently CYXZ
                bbox_size=bounding_box[[0, 2, 3, 1]],
            )

            # Reduce size
            crop_3d = image * 255
            crop_3d = crop_3d.astype(np.uint8)

            # Transpose to CZYX for saving
            crop_3d = transforms.transpose_to_dims(crop_3d, "CYXZ", "CZYX")

            # Save to OME-TIFF
            with OmeTiffWriter(cell_image_3d_save_path, overwrite_file=True) as writer:
                writer.save(
                    crop_3d,
                    dimension_order="CZYX",
                    channel_names=standardized_image.get_channel_names(),
                    pixels_physical_size=standardized_image.get_physical_pixel_size(),
                )

            # Generate 2d image projections
            # Crop raw channels using segmentations
            image = image_utils.crop_raw_channels_with_segmentation(image, channels)

            # Transpose to CZYX for projections
            image = transforms.transpose_to_dims(image, "CYXZ", "CZYX")

            # Select the DNA, Membrane, and Structure channels
            image = image[
                [
                    channels.index(target)
                    for target in [Channels.DNA, Channels.Membrane, Channels.Structure]
                ]
            ]

            # Set RGB colors
            # This will set:
            # DNA to Blue
            # Membrane to Red
            # Structure to Green
            colors = [[0, 0, 1], [1, 0, 0], [0, 1, 0]]

            # Get all axes projection image
            all_proj = proc.imgtoprojection(
                image,
                proj_all=True,
                proj_method=projection_method,
                local_adjust=False,
                global_adjust=True,
                colors=colors,
            )

            # Convert to YXC for PNG writing
            all_proj = transforms.transpose_to_dims(all_proj, "CYX", "YXC")

            # Drop size to uint8
            all_proj = all_proj.astype(np.uint8)

            # Save to PNG

            imwrite(cell_image_2d_all_proj_save_path, all_proj)

            # Get YX axes projection image
            yx_proj = proc.imgtoprojection(
                image,
                proj_all=False,
                proj_method=projection_method,
                local_adjust=False,
                global_adjust=True,
                colors=colors,
            )

            # Convert to YXC for PNG writing
            yx_proj = transforms.transpose_to_dims(yx_proj, "CYX", "YXC")

            # Drop size to uint8
            yx_proj = yx_proj.astype(np.uint8)

            # Save to PNG
            imwrite(cell_image_2d_yx_proj_save_path, yx_proj)

            log.info(f"Completed single cell image generation for CellId: {row.CellId}")

            # Return ready to save image
            return CellImagesResult(
                row.CellId,
                cell_image_3d_save_path,
                cell_image_2d_all_proj_save_path,
                cell_image_2d_yx_proj_save_path,
            )

        # Catch and return error
        except Exception as e:
            log.info(
                f"Failed single cell image generation for CellId: {row.CellId}. "
                "Error: {e}"
            )
            return CellImagesError(row.CellId, str(e))

    @log_run_params
    def run(
        self,
        dataset: Union[str, Path, pd.DataFrame, dd.DataFrame],
        cell_ceiling_adjustment: int = 7,
        bounding_box_percentile: float = 95.0,
        projection_method: str = "max",
        distributed_executor_address: Optional[str] = None,
        batch_size: Optional[int] = None,
        overwrite: bool = False,
        **kwargs,
    ):
        """
        Provided a dataset of cell features and standardized FOV images, generate 3D
        single cell crops and 2D projections.

        Parameters
        ----------
        dataset: Union[str, Path, pd.DataFrame, dd.DataFrame]
            The primary cell dataset to generate 3D single cell images for.

            **Required dataset columns:** *["CellId", "StandardizedFOVPath",
            "CellFeaturesPath"]*

        cell_ceiling_adjustment: int
            The adjust to use for raising the cell shape ceiling. If <= 0, this will be
            ignored and cell data will be selected but not adjusted.
            Default: 7

        bounding_box_percentile: float
            A float used to generate the actual bounding box for all cells by finding
            provided percentile of all cell image sizes.
            Default: 95.0

        projection_method: str
            The method to use for generating the flat projection.
            Default: max

            More details:
            https://allencellmodeling.github.io/aicsimageprocessing/aicsimageprocessing.html#aicsimageprocessing.imgToProjection.imgtoprojection

        distributed_executor_address: Optional[str]
            An optional executor address to pass to some computation engine.
            Default: None

        batch_size: Optional[int]
            An optional batch size to process n features at a time.
            Default: None (Process all at once)

        overwrite: bool
            If this step has already partially or completely run, should it overwrite
            the previous files or not.
            Default: False (Do not overwrite or regenerate files)

        Returns
        -------
        manifest_save_path: Path
            Path to the produced manifest with the various cell image path fields added.
        """
        # Handle dataset provided as string or path
        if isinstance(dataset, (str, Path)):
            dataset = Path(dataset).expanduser().resolve(strict=True)

            # Read dataset
            dataset = pd.read_csv(dataset)

        # Check dataset and manifest have required fields
        dataset_utils.check_required_fields(
            dataset=dataset, required_fields=REQUIRED_DATASET_FIELDS,
        )

        # Create save directories
        cell_images_3d_dir = self.step_local_staging_dir / "cell_images_3d"
        cell_images_2d_all_proj_dir = (
            self.step_local_staging_dir / "cell_images_2d_all_proj"
        )
        cell_images_2d_yx_proj_dir = (
            self.step_local_staging_dir / "cell_images_2d_yx_proj"
        )
        cell_images_3d_dir.mkdir(exist_ok=True)
        cell_images_2d_all_proj_dir.mkdir(exist_ok=True)
        cell_images_2d_yx_proj_dir.mkdir(exist_ok=True)

        # Process each row
        with DistributedHandler(distributed_executor_address) as handler:
            # Start processing
            bbox_results = handler.batched_map(
                self._get_registered_image_size,
                # Convert dataframe iterrows into two lists of items to iterate over
                # One list will be row index
                # One list will be the pandas series of every row
                *zip(*list(dataset.iterrows())),
                batch_size=batch_size,
            )

            # Compute bounding box with percentile
            bbox_results = np.array(bbox_results)
            bounding_box = np.percentile(bbox_results, bounding_box_percentile, axis=0)
            bounding_box = np.ceil(bounding_box)

            # Generate bounded arrays
            results = handler.batched_map(
                self._generate_single_cell_images,
                # Convert dataframe iterrows into two lists of items to iterate over
                # One list will be row index
                # One list will be the pandas series of every row
                *zip(*list(dataset.iterrows())),
                # Pass the other parameters as list of the same thing for each
                # mapped function call
                [cell_ceiling_adjustment for i in range(len(dataset))],
                [bounding_box for i in range(len(dataset))],
                [projection_method for i in range(len(dataset))],
                [cell_images_3d_dir for i in range(len(dataset))],
                [cell_images_2d_all_proj_dir for i in range(len(dataset))],
                [cell_images_2d_yx_proj_dir for i in range(len(dataset))],
                [overwrite for i in range(len(dataset))],
                batch_size=batch_size,
            )

        # Generate single cell images dataset rows
        single_cell_images_dataset = []
        errors = []
        for r in results:
            if isinstance(r, CellImagesResult):
                single_cell_images_dataset.append(
                    {
                        DatasetFields.CellId: r.cell_id,
                        DatasetFields.CellImage3DPath: r.path_3d,
                        DatasetFields.CellImage2DAllProjectionsPath: r.path_2d_all_proj,
                        DatasetFields.CellImage2DYXProjectionPath: r.path_2d_yx_proj,
                    }
                )
            else:
                errors.append({DatasetFields.CellId: r.cell_id, "Error": r.error})

        # Convert features paths rows to dataframe
        single_cell_images_dataset = pd.DataFrame(single_cell_images_dataset)

        # Drop the various single cell image columns if they already exist
        drop_columns = []
        if DatasetFields.CellImage3DPath in dataset.columns:
            drop_columns.append(DatasetFields.CellImage3DPath)
        if DatasetFields.CellImage2DAllProjectionsPath in dataset.columns:
            drop_columns.append(DatasetFields.CellImage2DAllProjectionsPath)
        if DatasetFields.CellImage2DYXProjectionPath in dataset.columns:
            drop_columns.append(DatasetFields.CellImage2DYXProjectionPath)

        dataset = dataset.drop(columns=drop_columns)

        # Join original dataset to the fov paths
        self.manifest = dataset.merge(
            single_cell_images_dataset, on=DatasetFields.CellId
        )

        # Save manifest to CSV
        manifest_save_path = self.step_local_staging_dir / "manifest.csv"
        self.manifest.to_csv(manifest_save_path, index=False)

        # Save errored cells to JSON
        with open(self.step_local_staging_dir / "errors.json", "w") as write_out:
            json.dump(errors, write_out)

        return manifest_save_path



# -*- coding: utf-8 -*-

from .single_cell_features import SingleCellFeatures
from .single_cell_images import SingleCellImages
from .standardize_fov_array import StandardizeFOVArray
from .make_diagnostic_sheet import MakeDiagnosticSheet

__all__ = [
    "SingleCellFeatures",
    "StandardizeFOVArray",
    "SingleCellImages",
    "MakeDiagnosticSheet",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import List, Union

import dask.dataframe as dd
import pandas as pd

###############################################################################


class MissingDataError(Exception):
    def __init__(
        self, dataset: Union[pd.DataFrame, dd.DataFrame], missing_fields: List[str]
    ):
        # Run base exception init
        super().__init__()

        # Store params for display
        self.dataset = dataset
        self.missing_fields = missing_fields

    def __str__(self):
        return (
            f"Dataset provided does not have the required columns for this operation. "
            f"Missing fields: {self.missing_fields} "
        )
